"""
Functionality for Gibbs sampler to generate posterior samples
from model from James Greiner, D. and Quinn, K.M., 2009.
R× C ecological inference: bounds, correlations, flexibility
and transparency of assumptions. Journal of the Royal Statistical
Society: Series A (Statistics in Society), 172(1), pp.67-81.
"""

import random
import scipy.stats as st
import numpy as np
import arviz as az
from tqdm import tqdm
from numba import njit, prange
from pyei.distribution_utils import non_central_hypergeometric_sample

__all__ = ["pyei_greiner_quinn_sample", "greiner_quinn_gibbs_sample"]


def pyei_greiner_quinn_sample(  # pylint: disable=too-many-locals
    group_fractions,
    votes_fractions,
    precinct_pops,
    num_samples=None,
    burnin=None,
    nu_0=None,
    psi_0=None,
    k_0_inv=None,
    mu_0=None,
    gamma=0.1,
):
    """
    Converts pyei inputs to counts for gq, sets default hyperparams,
    runs sampler, returns samples as an arviz InferenceData object

    Parameters:
    -----------
    group_fractions :   r x p (p =#precincts = num_precincts) matrix giving demographic
            information as the fraction of precinct_pop in the demographic group for each
            of p precincts and r demographic groups (sometimes denoted X)
    votes_fractions  :  c x p giving the fraction of each precinct_pop that votes
        for each of c candidates (sometimes denoted T)
    precinct_pops   :   Length-p array of ints giving size of each precinct population
                        of interest (e.g. voting population) (someteimes denoted N)
    num_samples: int (optional)
        number of MCMC samples to draw using the Gibbs sampler
    nu_0: int (optional)
        hyperparameter (df) scalar
        roughly interpretable as number of pseudodistricts for prior
    psi_0: ndarray (optional)
        hyperparameter (scale) - square matrix r * (c - 1) x r * (c - 1)
    k_0_inv: (optional)
        hyperparameter - square matrix r * (c - 1) x r * (c - 1)
        Precision for prior distribution over mu, the mean of the distribution of omega
        mu | mu_0, k_0 ~ N(mu_0, k_0) = N(mu_0, k_0_inv^{-1})
        omega | mu, Sigma ~ N(mu, Sigma)
    mu_0: ndarray (optional)
        vector of length r * (c - 1)
        Governs prior distribution over mu, the mean of the distribution of omega
        mu | mu_0, k_0 ~ N(mu_0, k_0)
        omega | mu, Sigma ~ N(mu, Sigma)
    gamma: float (optional)
        parameter of proposal dist in the Metropolis step for sampling theta
    burnin: int
        the number of samples at the start of the MCMC chain to be discarded

    Returns:
    --------
    idata: arviz.InverenceData
        An InferenceData object containing the posterior samples
        obtained via the Gibbs sampler

    Notes:
    -----
    This function converts group_fraction and votes_fraction to counts
    using the precinct_pops in the following manner: first,
    the fractions are multiplied by the precinct population and each rounded
    if the sum of the counts obtained by rounding don't match the precinct
    populations, the difference is subtracted from a *random* group or column
    """
    if num_samples is None:
        num_samples = 500_000
    if burnin is None:
        burnin = 100_000

    # covert fractions to counts
    # and make counts match precinct_pops (for example)
    # if a mismatch, add or subtract from a random group/candidate
    num_groups = group_fractions.shape[0]
    num_candidates = votes_fractions.shape[0]

    group_counts = np.round(group_fractions * precinct_pops)
    vote_counts = np.round(votes_fractions * precinct_pops)

    group_diff = group_counts.sum(axis=0) - precinct_pops
    for idx_of_mismatch in np.where(group_diff != 0):
        group_to_adjust = random.randint(0, num_groups - 1)
        group_counts[group_to_adjust, idx_of_mismatch] -= group_diff[idx_of_mismatch]

    vote_diff = vote_counts.sum(axis=0) - precinct_pops
    for idx_of_mismatch in np.where(vote_diff != 0):
        candidate_to_adjust = random.randint(0, num_candidates - 1)
        vote_counts[candidate_to_adjust, idx_of_mismatch] -= vote_diff[idx_of_mismatch]

    group_counts = group_counts.T
    vote_counts = vote_counts.T

    # Set any unspecified hyperparameters to default values
    if nu_0 is None:
        nu_0 = 10
    if psi_0 is None:
        psi_0 = (
            1 / 10 * np.identity(num_groups * (num_candidates - 1))
        )  # relates to prior precision
    if k_0_inv is None:
        k_0_inv = 1 / (0.5) * np.identity(num_groups * (num_candidates - 1))  # same size as Sigma
    if mu_0 is None:
        votes_all_precincts = vote_counts.sum(axis=0)
        log_ratio_support = np.log(
            votes_all_precincts[0:-1] / votes_all_precincts[-1]
        )  # assume each group supports according to the total vote split
        mu_0 = np.tile(log_ratio_support, num_groups)  # shape r * (c-1)

    samples = greiner_quinn_gibbs_sample(
        group_counts,
        vote_counts,
        num_samples,
        nu_0,
        psi_0,
        k_0_inv,
        mu_0,
        gamma=gamma,
        burnin=burnin,
    )
    # convert to InferenceData object
    for var_name in samples.keys():  # pylint: disable=consider-iterating-dictionary
        samples[var_name] = np.expand_dims(samples[var_name], 0)  # add an axis for chain
    samples["b"] = samples.pop("theta")  # changing variable name for vote prefernces to match pyei
    idata = az.from_dict(samples)

    return idata


def greiner_quinn_gibbs_sample(  # pylint: disable=too-many-locals
    group_counts, vote_counts, num_samples, nu_0, psi_0, k_0_inv, mu_0, gamma=0.1, burnin=0
):
    """
    group_counts: ndarray
        num_precincts x r gives number of people for each of r groups
        in num_precincts precincts
    vote_counts: ndarray
        num_precincts x c gives number of votes for each of c candidates
        in num_precincts precincts
    num_samples: int
        number of MCMC samples to draw using the Gibbs sampler
    nu_0: int
        hyperparameter (df) scalar
        roughly interpretable as number of pseudodistricts for prior
    psi_0: ndarray
        hyperparameter (scale) - square matrix r * (c - 1) x r * (c - 1)
    k_0_inv:
        hyperparameter - square matrix r * (c - 1) x r * (c - 1)
        Precision for prior distribution over mu, the mean of the distribution of omega
        mu | mu_0, k_0 ~ N(mu_0, k_0) = N(mu_0, k_0_inv^{-1})
        omega | mu, Sigma ~ N(mu, Sigma)
    mu_0: ndarray
        vector of length r * (c - 1)
        Governs prior distribution over mu, the mean of the distribution of omega
        mu | mu_0, k_0 ~ N(mu_0, k_0)
        omega | mu, Sigma ~ N(mu, Sigma)
    gamma: float
        parameter of proposal dist in the Metropolis step for sampling theta
    burnin: int
        the number of samples at the start of the MCMC chain to be discarded

    Note:
    -----
    Variable names follow those Greiner and Quinn, R x C ecological inference:
    bounds, correlations, flexibility and transparency of assumptions (2009)
    """
    # TODO: gamma set in initial runs?

    num_precincts = group_counts.shape[0]
    if np.all(num_precincts != vote_counts.shape[0]):
        raise ValueError(
            "group_counts and vote_counts must both have first dim of length num_precincts"
        )
    if burnin >= num_samples:
        raise ValueError(f"num_samples is only {num_samples} but burn-in is {burnin}")

    precinct_pops = vote_counts.sum(axis=1)
    _, num_candidates = vote_counts.shape
    _, num_groups = group_counts.shape

    # set initial values
    mu_samp = mu_0
    Sigma_samp = st.invwishart.rvs(df=nu_0, scale=psi_0)
    omega_samp = np.zeros((num_precincts, num_groups * (num_candidates - 1)))
    print(num_groups, num_candidates)
    theta_samp = omega_to_theta(omega_samp, num_groups, num_candidates)
    internal_cell_counts_samp = get_initial_internal_count_sample(
        group_counts, vote_counts, precinct_pops
    )

    # variables for storing all samples
    internal_cell_counts_samples = np.empty(
        (num_samples - burnin, num_precincts, num_groups, num_candidates)
    )
    theta_samples = np.empty((num_samples - burnin, num_precincts, num_groups, num_candidates))
    mu_samples = np.empty((num_samples - burnin, num_groups * (num_candidates - 1)))
    sigma_samples = np.empty(
        (num_samples - burnin, num_groups * (num_candidates - 1), num_groups * (num_candidates - 1))
    )

    for i in tqdm(range(num_samples)):
        # (a) sample internal cell counts
        internal_cell_counts_samp = sample_internal_cell_counts(
            theta_samp, internal_cell_counts_samp
        )

        # (b) sample theta using Metropolis-Hastings
        theta_samp = sample_theta(
            internal_cell_counts_samp, theta_samp, omega_samp, mu_samp, Sigma_samp, gamma
        )

        omega_samp = theta_to_omega(theta_samp)
        # TODO: IS THE NEXT LINE UNNECESSARY?
        omega_samp = omega_samp.reshape((num_precincts, num_groups * (num_candidates - 1)))

        # (c) sample mu and sigma given omega (or, equivalently, given theta)
        mu_samp = sample_mu(omega_samp, Sigma_samp, k_0_inv, mu_0, num_precincts)
        Sigma_samp = sample_Sigma(omega_samp, mu_samp, nu_0, psi_0)

        # save samples after burn-in
        if i >= burnin:
            internal_cell_counts_samples[i - burnin, :, :, :] = internal_cell_counts_samp
            theta_samples[i - burnin, :, :, :] = theta_samp
            mu_samples[i - burnin, :] = mu_samp
            sigma_samples[i - burnin, :, :] = Sigma_samp

    return {
        "theta": theta_samples,
        "counts": internal_cell_counts_samples,
        "mu": mu_samples,
        "Sigma": sigma_samples,
    }


def sample_Sigma(omega, mu, nu_0, psi_0):
    """
    Parameters:
    -----------
    omega: ndarray
        num_precints x (r * (c - 1))
        A transformation of the voting preferences given by theta,
        omega(i, r', c') = log(theta(i,r',c')/theta(i,r, num_candidates)),
    mu: array
        vector of length r * (c - 1)
        mean of the normal distribution that governs omega
        omega | mu, Sigma ~ N(mu, Sigma)
        where omega(i, r', c') = log(theta(i,r',c')/theta(i,r, c'')),
        theta are the voting preferences of each group and candidate in
        each precinct, and c'' is some reference column/candidate
        (here the last column)
    nu_0: float
        hyperparameter (deg of freedom) - scalar
    psi_0: n
        darray:hyperparameter (scale)
        square matrix r * (c - 1) x r * (c - 1)

    Returns:
    --------
    Sigma: ndarray
        square matrix r * (c - 1) x r * (c - 1), the covariance of omega
    """
    num_precincts = omega.shape[0]
    nu_n = nu_0 + num_precincts
    psi_n = psi_0 + ((omega - mu) @ (omega - mu).T).sum()  # sum over precincts

    Sigma = st.invwishart.rvs(nu_n, psi_n)
    return Sigma


def sample_mu(omega, Sigma, k_0_inv, mu_0, num_precincts):
    """
    omega: num_precints x (r * (c - 1))
    Sigma: ndarray
        square matrix r * (c - 1) x r * (c - 1), the covariance of omega
    k_0_inv is hyperparameter - square matrix r * (c - 1) x r * (c - 1)
    mu_0 : array
        vector of length r * (c - 1)
        mean of the prior distribution for mu, mu the mean of of the normal
        distribution that governs omega
        mu | mu_0, k_0 ~ N(mu_0, k_0)
        omega | mu, Sigma ~ N(mu, Sigma)
        where omega(i, r', c') = log(theta(i,r',c')/theta(i,r, c'')),
        theta are the voting preferences of each group and candidate in
        each precinct, and c'' is some reference column/candidate
        (here the last column)
    num_precincts: int
        The number of precincts

    Returns:
    mu: a vector of length r * (c-1)
    """

    Sigma_inv = np.linalg.inv(Sigma)
    mean_omega = np.mean(omega, axis=0)
    mu_n = (np.linalg.inv(k_0_inv + num_precincts * Sigma_inv)) @ (
        k_0_inv @ mu_0.T + num_precincts * Sigma_inv @ mean_omega
    )
    Sigma_n_inv = k_0_inv + num_precincts * Sigma_inv
    mu = st.multivariate_normal.rvs(mean=mu_n, cov=np.linalg.inv(Sigma_n_inv))

    return mu


def proposal_dist_generate_sample(mu_samp, Sigma_samp, gamma, num_precincts, deg_freedom=4):
    """
    Grainer and Quinn use a t_4(mu_t, gamma * Sigma) proposal dist with gamma
    set during inital runs - this generates an omega sample, which they transform
    back to theta space

    Returns:
    --------
    omega_proposed:n num_precincts * (r * (c-1)
    """
    omega_proposed = st.multivariate_t.rvs(
        mu_samp, gamma * Sigma_samp, df=deg_freedom, size=num_precincts
    )
    return omega_proposed


def log_unnormalized_pdf(theta, omega, internal_cell_counts_samp, mu_samp, Sigma_samp, tol=0.01):
    """pdf proportional t to the product of lines (4)-(6) and (10) and (11) in G&Q
    0 if thetas don't sum to 1 across rows

    Sigma_samp: ndarray
        square matrix r * (c - 1) x r * (c - 1), a sample of the covariance of omega
    tol: float
        how far can the sum of theta be from 1
    theta: ndarray
        num_precints x r x c
    """

    # Lines (10) - (11) are this check
    if not np.all(abs(theta.sum(axis=2) - 1) < tol):  # CHECK IF THETAS SUM TO 1 across rows
        return -np.inf  # if not, prob is zero, so log_prob is -inf

    else:
        line_4_and_6 = np.apply_over_axes(
            np.sum, (internal_cell_counts_samp - 1) * np.log(theta), [1, 2]
        ).flatten()  # sum over rows and columns
        line_5 = -0.5 * np.linalg.det(Sigma_samp) - 0.5 * (
            (omega - mu_samp) @ np.linalg.inv(Sigma_samp) * (omega - mu_samp)
        ).sum(axis=1)
        return (line_4_and_6 + line_5).sum()  # sum over precincts


def theta_to_omega(theta):
    """
    Parameters:
    -----------
    theta: ndarray
        num_precints x r x c

    Returns:
    --------
    omega: ndarray
        num_precincts x r x (c-1)
    """
    return np.log(theta[:, :, :-1]) - np.log(theta[:, :, [-1]])


def sample_theta(internal_cell_counts_samp, theta_prev, omega_prev, mu_samp, Sigma_samp, gamma):
    """
    Use a Metropolis-Hastings step to sample theta

    internal_cell_counts_samp: ndarray
    theta_prev: ndarray
    omega_prev: ndarray
    mu_samp:
    Sigma_samp: ndarray
        square matrix r * (c - 1) x r * (c - 1), a sample of the covariance of omega
    gamma: float
        parameter of proposal dist in the Metropolis step

    """
    # MH
    num_precincts, r, c = internal_cell_counts_samp.shape

    omega_proposed = proposal_dist_generate_sample(
        mu_samp, Sigma_samp, gamma, num_precincts, deg_freedom=4
    )

    theta_proposed = omega_to_theta(omega_proposed, r, c)

    log_unnormalized_pdf_proposed = log_unnormalized_pdf(
        theta_proposed, omega_proposed, internal_cell_counts_samp, mu_samp, Sigma_samp
    )
    log_unnormalized_pdf_prev = log_unnormalized_pdf(
        theta_prev, omega_prev, internal_cell_counts_samp, mu_samp, Sigma_samp
    )

    # calculate acceptance prbability
    log_acc_prob = log_unnormalized_pdf_proposed - log_unnormalized_pdf_prev
    log_acc_prob = np.min([log_acc_prob, 0])
    acc_prob = np.exp(log_acc_prob)

    unif_samp = st.uniform.rvs(size=1)
    if unif_samp < acc_prob:
        return theta_proposed
    else:
        return theta_prev


def omega_to_theta(omega, r, c):
    """
    theta: num_precints x r x c
    omega: num_precincts x (r * (c-1))

    Note:
    -----
    SIDE EFFECT: RESHAPES OMEGA
    """
    num_precincts = omega.shape[0]
    omega = omega.reshape(num_precincts, r, c - 1)
    theta_last = 1.0 / (np.exp(omega).sum(axis=2) + 1)  # num_precincts x r x 1
    theta_other = (np.exp(omega).T * theta_last.T).T  # num_precincts x r x c - 1
    theta_last_ext = np.expand_dims(theta_last, axis=2)
    return np.concatenate((theta_other, theta_last_ext), axis=2)


@njit(parallel=True)
def sample_internal_cell_counts(theta_samp, prev_internal_counts_samp):
    """
    group_counts: num_precincts x r
    vote_counts: num_precincts x c
    theta: num_precints x r x c
    prev_internal_counts_samp: num_precincts x r x c

    @TODO VECTORIZE
    """
    num_precincts, num_groups, num_candidates = prev_internal_counts_samp.shape

    for i in prange(num_precincts):  # pylint: disable=not-an-iterable
        for r in range(num_groups - 1):
            for r_prime in range(r + 1, num_groups):
                for c in range(num_candidates - 1):
                    for c_prime in range(c + 1, num_candidates):
                        n1 = (  # pylint: disable=invalid-name
                            prev_internal_counts_samp[i, r, c]
                            + prev_internal_counts_samp[i, r, c_prime]
                        )  # n1 gives row total in row r
                        n2 = (  # pylint: disable=invalid-name
                            prev_internal_counts_samp[i, r_prime, c]
                            + prev_internal_counts_samp[i, r_prime, c_prime]
                        )  # n2 row total in row r_prime
                        m1 = (  # pylint: disable=invalid-name
                            prev_internal_counts_samp[i, r, c]
                            + prev_internal_counts_samp[i, r_prime, c]
                        )  # m1 gives column total in column c
                        pi1 = theta_samp[i, r, c]
                        pi2 = theta_samp[i, r_prime, c]
                        psi = (pi1 * (1 - pi2)) / (pi2 * (1 - pi1))
                        r_c_count = non_central_hypergeometric_sample(
                            n1, n2, m1, psi
                        )  # sample for the r, c internal count

                        # update prev_internal counts in the 2 x 2 subarray
                        prev_internal_counts_samp[i, r, c] = r_c_count
                        prev_internal_counts_samp[i, r, c_prime] = n1 - r_c_count
                        prev_internal_counts_samp[i, r_prime, c] = m1 - r_c_count
                        prev_internal_counts_samp[i, r_prime, c_prime] = (
                            n2 - prev_internal_counts_samp[i, r_prime, c]
                        )
    return prev_internal_counts_samp


@njit(parallel=True, nopython=False)
def get_initial_internal_count_sample(group_counts, vote_counts, precinct_pops):
    """
    Sets an initial value of internal counts that is compatible with the
    observed vote and group counts

    Parameters:
    -----------
    group_counts: ndarray
        num_precincts x r. Give count of people in each group within each
        precinct.
    vote_counts: ndarray
        num_precincts x c. Gives count of votes for each candidate within each
        precinct
    precinct_pops: vector of length num_precincts

    Returns:
    --------
        internal_counts_samp: ndarray
            num_precincts x r x c
            Gives a set of "internal counts" in the table of results
            for each precinct - gives a set of counts for people
            within each group that voted for each candidates that
            is possible given the marginal sums
            vote_counts and group_counts

    Notes:
    ------
    Within each of rows up to row num_groups-1, for c=0,..num_candidates-1
    and samples according to a binomial with
    p=vote_counts[i, c] / precinct_pops[i],
    making sure not to exceed the remaining counts left to assign in the
    column and row.
    Keep track of remaining counts in each row and column and use that
    to fill the last row and column.
    """
    num_precincts, num_groups = group_counts.shape
    _, num_candidates = vote_counts.shape
    internal_counts_samp = np.empty((num_precincts, num_groups, num_candidates), dtype=np.int16)

    group_counts_remaining = group_counts.copy()
    vote_counts_remaining = vote_counts.copy()

    for i in prange(num_precincts):  # pylint: disable=not-an-iterable
        for r in range(num_groups - 1):
            for c in range(num_candidates - 1):
                count_for_binom = np.round(
                    group_counts[i, r]
                    * (vote_counts[i, c] + vote_counts[i, c + 1])
                    / precinct_pops[i]
                )
                prop_for_binom = vote_counts[i, c] / precinct_pops[i]
                samp = np.random.binomial(n=int(count_for_binom), p=int(prop_for_binom))

                samp = min(samp, vote_counts_remaining[i, c])
                samp = min(samp, group_counts_remaining[i, r])
                group_counts_remaining[i, r] = group_counts_remaining[i, r] - samp
                vote_counts_remaining[i, c] = vote_counts_remaining[i, c] - samp
                internal_counts_samp[i, r, c] = samp
            internal_counts_samp[i, r, num_candidates - 1] = group_counts_remaining[i, r]
            vote_counts_remaining[i, num_candidates - 1] = (
                vote_counts_remaining[i, num_candidates - 1]
                - internal_counts_samp[i, r, num_candidates - 1]
            )
        internal_counts_samp[i, num_groups - 1, :] = vote_counts_remaining[i, :]

    return internal_counts_samp



"""
Goodman's ecological regression
"""

from matplotlib import pyplot as plt
import numpy as np
import pymc as pm
from sklearn.linear_model import LinearRegression
import seaborn as sns
import seaborn.algorithms as sns_algo
import seaborn.utils as sns_utils

from .two_by_two import TwoByTwoEIBaseBayes

__all__ = ["GoodmansER", "GoodmansERBayes"]


class GoodmansER:
    """
    Fitting and plotting for Goodman's ER (with options for pop weighting)
    """

    def __init__(self, is_weighted_regression=False):
        """
        Parameters
        ----------
        is_weighted_regression: bool, optional
            Default is False. If true, weight precincts by population when
            performing the regression.
        """
        self.demographic_group_fraction = None
        self.vote_fraction = None
        self.demographic_group_fraction = None
        self.demographic_group_name = None
        self.candidate_name = None
        self.intercept_ = None
        self.slope_ = None
        self.voting_prefs_est_ = None
        self.voting_prefs_complement_est_ = None
        self.precinct_pops = None
        self.is_weighted_regression = is_weighted_regression

    def fit(
        self,
        group_fraction,
        vote_fraction,
        precinct_pops=None,
        demographic_group_name="given demographic group",
        candidate_name="given candidate",
    ):
        """Fit the linear model (use pop weights iff is_weighted_regression is true

        Parameters
        ----------
        group_fraction  :   Length-p (p=# of precincts) vector giving demographic
                            information (X) as the fraction of precinct_pop in
                            the demographic group of interest
        vote_fraction  :   Length p vector giving the fraction of each precinct_pop
                            that votes for the candidate of interest (T)
        precinct_pops   :   Length-p vector giving size of each precinct population
                            of interest (e.g. voting population) (N)
                            Used for weighting the regression if the GoodmansER
                            object has is_weighted_regression True
        demographic_group_name  :   Name of the demographic group of interest,
                                    where results are computed for the
                                    demographic group and its complement
        candidate_name          :   Name of the candidate whose support we
                                    want to analyze
        """
        self.demographic_group_fraction = group_fraction
        self.vote_fraction = vote_fraction
        self.precinct_pops = precinct_pops
        self.demographic_group_name = demographic_group_name
        self.candidate_name = candidate_name
        if self.is_weighted_regression:
            reg = LinearRegression().fit(
                group_fraction.reshape(-1, 1),
                vote_fraction,
                sample_weight=precinct_pops,
            )
        else:
            reg = LinearRegression().fit(group_fraction.reshape(-1, 1), vote_fraction)

        self.intercept_ = reg.intercept_
        self.slope_ = reg.coef_[0]
        self.voting_prefs_est_ = reg.predict(np.array([1]).reshape(-1, 1))[0]
        self.voting_prefs_complement_est_ = reg.intercept_
        return self

    def summary(self):
        """Return summary of results as string"""
        if self.is_weighted_regression:
            model_name = "Goodmans ER, weighted by population"
        else:
            model_name = "Goodmans ER"
        return f"""{model_name}
        Est. fraction of {self.demographic_group_name}
        voters who voted for {self.candidate_name} is
        {self.voting_prefs_est_:.3f}
        Est. fraction of non- {self.demographic_group_name}
        voters who voted for {self.candidate_name} is
        {self.voting_prefs_complement_est_:.3f}
        """

    def plot(
        self,
        line_kws=None,
        scatter_kws=None,
        **sns_regplot_args,
    ):
        """Plot the linear regression with 95% confidence interval

        Notes:
        ------
        Can pass additional plot arguments through to seaborn.regplot, e.g.
        to change the line color:
            line_kws=dict(color="red")
            scatter_kws={"s": 50}
        """
        fig, ax = plt.subplots()
        ax.axis("square")
        ax.grid(visible=True, which="major")
        ax.set_ylim((0, 1))
        ax.set_xlim((0, 1))
        ax.set_xlabel(f"Fraction in group {self.demographic_group_name}")
        ax.set_ylabel(f"Fraction voting for {self.candidate_name}")

        if self.is_weighted_regression:
            if line_kws is None:
                line_kws = {}
            line_kws.setdefault("linewidth", 2.5)
            if scatter_kws is None:
                scatter_kws = {}
            scatter_kws.setdefault("s", 50)
            scatter_kws.setdefault("linewidths", 0)
            scatter_kws.setdefault("alpha", 0.8)
            sns.lineplot(
                x=[0, 1], y=[self.intercept_, self.intercept_ + self.slope_], ax=ax, **line_kws
            )
            sns.scatterplot(
                x=self.demographic_group_fraction, y=self.vote_fraction, ax=ax, **scatter_kws
            )
            xgrid = np.linspace(0, 1, 101)

            def fit_fast(xgrid, x, y, w):
                """
                Modifying this function from sns to accomodate weighted regression
                in CI computation
                """

                def weighted_reg_func(_x, _y, _w):
                    """Low-level regression and prediction using linear algebra."""
                    _w_sqrt = np.sqrt(_w)
                    x_weighted = np.diag(_w_sqrt).dot(_x)
                    y_weighted = np.diag(_w_sqrt).dot(_y)
                    return np.linalg.pinv(x_weighted).dot(y_weighted)

                x_with_ones = np.c_[np.ones(len(x)), x]
                grid = np.c_[np.ones(len(xgrid)), xgrid]  # append ones for intercept
                yhat = grid.dot(weighted_reg_func(x_with_ones, y, w))
                beta_boots = sns_algo.bootstrap(
                    x_with_ones, y, w, func=weighted_reg_func, n_boot=1000, units=None, seed=None
                ).T
                yhat_boots = grid.dot(beta_boots).T
                return yhat, yhat_boots

            _, yhat_boots = fit_fast(
                xgrid,
                self.demographic_group_fraction,
                self.vote_fraction,
                self.precinct_pops,
            )
            # CIs at each grid point
            err_bands = sns_utils.ci(yhat_boots, which=95, axis=0)
            ax.fill_between(
                xgrid, *err_bands, alpha=0.15, color=plt.gca().lines[-1].get_color()
            )  # match color of plot
        else:
            sns.regplot(
                x=self.demographic_group_fraction,
                y=self.vote_fraction,
                ax=ax,
                ci=95,
                truncate=False,
                **sns_regplot_args,
            )
        return fig, ax


class GoodmansERBayes(TwoByTwoEIBaseBayes):
    """Bayesian ecological regression with uniform prior over the voting preferences
    Generate samples from the posterior.
    """

    def __init__(
        self,
        model_name="goodman_er_bayes",
        weighted_by_pop=False,
        **additional_model_params,
    ):
        """
        Optional arguments:
        model_name: str
            Default is "goodman_er_bayes"
        weighted_by_pop: bool
            Default is False. If true, weight precincts by population when
            performing the regression.
        additional_model_parameters:
            Any hyperparameters for model
        """
        # TODO if no other model name is applicable here, remove need for model_name argument
        super().__init__(model_name, **additional_model_params)
        self.weighted_by_pop = weighted_by_pop

    def fit(
        self,
        group_fraction,
        votes_fraction,
        precinct_pops=None,
        demographic_group_name="given demographic group",
        candidate_name="given candidate",
        target_accept=0.9,
        tune=1000,
        **other_sampling_args,
    ):
        """Fit a bayesian er modeling via sampling.

        Parameters
        ----------
        group_fraction  :   Length-p (p=# of precincts) vector giving demographic
                            information (X) as the fraction of precinct_pop in
                            the demographic group of interest
        votes_fraction  :   Length p vector giving the fraction of each precinct_pop
                            that votes for the candidate of interest (T)
        precinct_pops   :   Length-p vector giving size of each precinct population
                            of interest (e.g. voting population) (N)
        demographic_group_name  :   Name of the demographic group of interest,
                                    where results are computed for the
                                    demographic group and its complement
        candidate_name          :   Name of the candidate whose support we
                                    want to analyze
        precinct_names          :   Length p vector giving the string names
                                    for each precinct.
        target_accept : float, optional
            Default=.99 Strictly between zero and 1 (should be close to 1). Passed to pymc's
            sampling.sample
        tune : int, optional
            Default=1500, passed to pymc's sampling.sample
        other_sampling_args :
            For to pymc's sampling.sample
            https://docs.pymc.io/api/inference.html
        """
        self.precinct_pops = precinct_pops
        self.demographic_group_name = demographic_group_name
        self.candidate_name = candidate_name
        self.demographic_group_fraction = group_fraction
        self.votes_fraction = votes_fraction

        if self.weighted_by_pop:
            model_function = goodmans_er_bayes_pop_weighted_model
            self.additional_model_params["precinct_pops"] = precinct_pops
        else:
            model_function = goodmans_er_bayes_model
        self.sim_model = model_function(
            group_fraction, votes_fraction, **self.additional_model_params
        )

        with self.sim_model:  # pylint: disable=not-context-manager
            self.sim_trace = pm.sample(
                1000, tune=tune, target_accept=target_accept, **other_sampling_args
            )

        self.calculate_sampled_voting_prefs()
        super().calculate_summary()

    def calculate_sampled_voting_prefs(self):
        """Sets sampled_voting_prefs"""
        # obtain samples of the districtwide proportion of each demog. group voting for candidate
        self.sampled_voting_prefs[0] = (
            self.sim_trace["posterior"]["b_1"].stack(all_draws=["chain", "draw"]).values.T
        )
        # sampled voted prefs across precincts
        self.sampled_voting_prefs[1] = (
            self.sim_trace["posterior"]["b_2"].stack(all_draws=["chain", "draw"]).values.T
        )
        # sampled voted prefs across precincts

    def compute_credible_int_for_line(self, x_vals=np.linspace(0, 1, 100)):
        """Computes regression line (mean) and 95% central credible interval for
        the mean line at each of the specified x values(x_vals)

        Parameters
        ----------
        x_vals : numpy array, optional
            Default: np.linspace(0, 1, 100). 1-dimensional numpy array of values between
            0 and 1, for which the function should compute the mean line and 95%
            credible interval. Each element of x_vals represents the fraction of the population
            that is in the demographic group of interest (values for X in the notation of King '97)

        """
        lower_bounds = np.empty_like(x_vals)
        upper_bounds = np.empty_like(x_vals)
        means = np.empty_like(x_vals)
        for idx, x in enumerate(x_vals):
            mean_samples = (
                self.sampled_voting_prefs[1]
                + (self.sampled_voting_prefs[0] - self.sampled_voting_prefs[1]) * x
            )
            percentiles = np.percentile(mean_samples, [2.5, 97.5])
            lower_bounds[idx] = percentiles[0]
            upper_bounds[idx] = percentiles[1]
            means[idx] = mean_samples.mean()

        return x_vals, means, lower_bounds, upper_bounds

    def plot(self, scatter_kws=None, line_kws=None):
        """Plot regression line of votes_fraction vs. group_fraction, with scatter plot and
        equal-tailed 95% credible interval for the line"
        Parameters:
        -----------
        scatter_kws : dict (None)
            Keyword arguments to be passed to matplotlib.Axes.scatter
        line_kws : dict (None)
            Keyword arguments to be passed to matplotlib.Axes.plot.
            Note that the color of the ccredible interval shading is set to match
            the color of the line itself (but the shading has higher alpha)

        Notes:
        ------
        Examples of additional kwargs. Scatter_colors is list of colors of length num_precincts
        scatter_kws={"c": scatter_colors, "color": None, "s": 20},
        line_kws={"color":"black", "lw": 1}
        """
        # TODO: consider renaming these plots for goodman, to disambiguate with TwoByTwoEI.plot()
        # TODO: accept axis argument
        x_vals, means, lower_bounds, upper_bounds = self.compute_credible_int_for_line()
        _, ax = plt.subplots()

        if scatter_kws is None:
            scatter_kws = {}
        if line_kws is None:
            line_kws = {}
        scatter_kws.setdefault("color", "steelblue")
        scatter_kws.setdefault("alpha", 0.8)
        line_kws.setdefault("color", "steelblue")

        ax.axis("square")
        ax.set_xlabel(f"Fraction in group {self.demographic_group_name}")
        ax.set_ylabel(f"Fraction voting for {self.candidate_name}")
        ax.scatter(
            self.demographic_group_fraction,
            self.votes_fraction,
            **scatter_kws,
        )
        ax.plot(x_vals, means, **line_kws)
        ax.fill_between(x_vals, lower_bounds, upper_bounds, alpha=0.2, color=line_kws["color"])
        ax.grid()
        ax.set_xlim((0, 1))
        ax.set_ylim((0, 1))
        return ax


def goodmans_er_bayes_model(group_fraction, votes_fraction, sigma=1):
    """Ecological regression with uniform priors over voting prefs b_1, b_2,
    constraining them to be between zero and 1

    Parameters
    ----------
    group_fraction  :   Length-p (p=# of precincts) vector giving demographic
                            information (X) as the fraction of precinct_pop in
                            the demographic group of interest
    votes_fraction  :   Length p vector giving the fraction of each precinct_pop
                            that votes for the candidate of interest (T)

    Returns
    -------
    bayes_er_model: a pymc3 model
    """

    with pm.Model() as bayes_er_model:
        b_1 = pm.Uniform("b_1")
        b_2 = pm.Uniform("b_2")

        eps = pm.HalfNormal("eps", sigma=sigma)

        pm.Normal(
            "votes_count_obs",
            b_2 + (b_1 - b_2) * group_fraction,
            sigma=eps,
            observed=votes_fraction,
        )

    return bayes_er_model


def goodmans_er_bayes_pop_weighted_model(group_fraction, votes_fraction, precinct_pops, sigma=1):
    """Ecological regression with variance of modeled vote fraction inversely proportional to
    precinct population.

    Uniform priors over voting prefs b_1, b_2 constrain them to be between 0 and 1

    Parameters
    ----------
    group_fraction  :   Length-p (p=# of precincts) vector giving demographic
                            information (X) as the fraction of precinct_pop in
                            the demographic group of interest
    votes_fraction  :   Length p vector giving the fraction of each precinct_pop
                            that votes for the candidate of interest (T)
    precinct_pops   :   Length-p vector giving size of each precinct population
                            of interest (e.g. voting population) (N)

    Returns
    -------
    bayes_er_model: a pymc3 model
    """

    mean_precinct_pop = precinct_pops.mean()
    with pm.Model() as bayes_er_model:
        b_1 = pm.Uniform("b_1")
        b_2 = pm.Uniform("b_2")

        eps = pm.HalfNormal("eps", sigma=sigma)

        pm.Normal(
            "votes_count_obs",
            b_2 + (b_1 - b_2) * group_fraction,
            sigma=eps * pm.math.sqrt(mean_precinct_pop / precinct_pops),
            observed=votes_fraction,
        )

    return bayes_er_model



"""
Functions that return pymc models for RowByColumnEI
"""

import pymc as pm
import pytensor.tensor as at
import numpy as np

__all__ = ["ei_multinom_dirichlet_modified", "ei_multinom_dirichlet"]


def ei_multinom_dirichlet(group_fractions, votes_fractions, precinct_pops, lmbda1=4, lmbda2=2):
    """
    An implementation of the r x c dirichlet/multinomial EI model

    Parameters:
    -----------
    group_fractions: r x num_precincts  matrix giving demographic information
        as the fraction of precinct_pop in the demographic group of interest for each of
        p precincts and r demographic groups (sometimes denoted X)
    votes_fractions: c x num_precincts matrix giving the fraction of each precinct_pop that votes
        for each of c candidates (sometimes denoted T)
    precinct_pops: Length-num_precincts vector giving size of each precinct population of interest
        (e.g. voting population) (sometimes denoted N)
    lmbda1: float parameter passed to the Gamma(lmbda, 1/lmbda2) distribution
    lmbda2: float parameter passed to the Gamma(lmbda, 1/lmbda2) distribution

    Returns
    -------
    model: A pymc3 model
    """

    num_precincts = len(precinct_pops)  # number of precincts
    num_rows = group_fractions.shape[0]  # number of demographic groups (r)
    num_cols = votes_fractions.shape[0]  # number of candidates or voting outcomes (c)

    # reshaping and rounding
    votes_count_obs = np.swapaxes(
        votes_fractions * precinct_pops, 0, 1
    ).round()  # num_precincts x r
    group_fractions_extended = np.expand_dims(group_fractions, axis=2)
    group_fractions_extended = np.repeat(group_fractions_extended, num_cols, axis=2)
    group_fractions_extended = np.swapaxes(group_fractions_extended, 0, 1)
    # num_precincts x r x c

    with pm.Model() as model:
        # TODO: are the prior conc_params what is in the literature? is it a good choice?
        # TODO: make b vs. beta naming consistent
        # conc_params = pm.Exponential("conc_params", lam=lmbda, shape=(num_rows, num_cols))
        conc_params = pm.Gamma(
            "conc_params", alpha=lmbda1, beta=1 / lmbda2, shape=(num_rows, num_cols)
        )  # chosen to match eiPack
        beta = pm.Dirichlet("b", a=conc_params, shape=(num_precincts, num_rows, num_cols))
        # num_precincts x r x c
        theta = (group_fractions_extended * beta).sum(axis=1)
        pm.Multinomial(
            "votes_count",
            n=precinct_pops,
            p=theta,
            observed=votes_count_obs,
            shape=(num_precincts, num_cols),
        )  # num_precincts x c
    return model


def ei_multinom_dirichlet_modified(
    group_fractions, votes_fractions, precinct_pops, pareto_scale=5, pareto_shape=1
):
    """
    An implementation of the r x c dirichlet/multinomial EI model with reparametrized hyperpriors

    Parameters:
    -----------
    group_fractions: r x num_precincts  matrix giving demographic information
        as the fraction of precinct_pop in the demographic group of interest for each of
        p precincts and r demographic groups (sometimes denoted X)
    votes_fractions: c x num_precincts matrix giving the fraction of each precinct_pop that votes
        for each of c candidates (sometimes denoted T)
    precinct_pops: Length-num_precincts vector giving size of each precinct population of interest
        (e.g. voting population) (sometimes denoted N)

    Returns
    -------
    model: A pymc3 model

    Notes
    -----
    Reparametrizing of the hyperpriors to give (hopefully) better geometry for sampling.
    Also gives intuitive interpretation of hyperparams as mean and counts
    """

    num_precincts = len(precinct_pops)  # number of precincts
    num_rows = group_fractions.shape[0]  # number of demographic groups (r)
    num_cols = votes_fractions.shape[0]  # number of candidates or voting outcomes (c)

    # reshaping and rounding
    votes_count_obs = np.swapaxes(
        votes_fractions * precinct_pops, 0, 1
    ).round()  # num_precincts x r
    group_fractions_extended = np.expand_dims(group_fractions, axis=2)
    group_fractions_extended = np.repeat(group_fractions_extended, num_cols, axis=2)
    group_fractions_extended = np.swapaxes(group_fractions_extended, 0, 1)
    # num_precincts x r x c

    with pm.Model() as model:
        # TODO: make b vs. beta naming consistent
        kappa = pm.Pareto("kappa", alpha=pareto_shape, m=pareto_scale, shape=num_rows)  # size r
        phi = pm.Dirichlet("phi", a=np.ones(num_cols), shape=(num_rows, num_cols))  # r x c
        phi_kappa = pm.Deterministic("phi_kappa", at.transpose(kappa * at.transpose(phi)))
        beta = pm.Dirichlet("b", a=phi_kappa, shape=(num_precincts, num_rows, num_cols))
        # num_precincts x r x c
        theta = (group_fractions_extended * beta).sum(axis=1)  # sum across num_rows
        pm.Multinomial(
            "votes_count",
            n=precinct_pops,
            p=theta,
            observed=votes_count_obs,
            shape=(num_precincts, num_cols),
        )  # num_precincts x c
    return model



"""
Utilities for RowByColumnEI
"""

import warnings


def check_dimensions_of_input(
    group_fractions,
    votes_fractions,
    precinct_pops,
    demographic_group_names,
    candidate_names,
    num_groups_and_num_candidates,
):
    """Checks shape of inputs and gives warnings or errors if there is a problem

    Required arguments:
    group_fractions :   r x p (p =#precincts = num_precicts) matrix giving demographic
        information as the fraction of precinct_pop in the demographic group for each
        of p precincts and r demographic groups (sometimes denoted X)
    votes_fractions  :  c x p giving the fraction of each precinct_pop that votes
        for each of c candidates (sometimes denoted T)
    precinct_pops   :   Length-p vector giving size of each precinct population
                        of interest (e.g. voting population) (someteimes denoted N)
    Optional arguments:
    demographic_group_names  :  Names of the r demographic group of interest,
                                where results are computed for the
                                demographic group and its complement
    candidate_names          :  Name of the c candidates or voting outcomes of interest

    """

    if demographic_group_names is not None:
        if len(demographic_group_names) != num_groups_and_num_candidates[0]:
            warnings.warn(
                """Length of demographic_groups_names should be equal to
            r = group_fractions.shape[0]. If not, plotting labels may be inaccurate.
            """
            )

    if candidate_names is not None:
        if len(candidate_names) != num_groups_and_num_candidates[1]:
            warnings.warn(
                """Length of candidate_names should be equal to
            c = votes_fractions.shape[0]. If not, plotting labels be inaccurate.
            """
            )

    print(f"Running {demographic_group_names} x {candidate_names} EI")
    print(f"r = {num_groups_and_num_candidates[0]} rows (demographic groups)")
    print(f"c = {num_groups_and_num_candidates[1]} columns (candidates or voting outcomes)")
    print(f"number of precincts = {len(precinct_pops)}")

    if len(precinct_pops) != votes_fractions.shape[1]:
        raise ValueError(
            """votes_fractions should have shape: c x num_precincts.
        In particular, it is required that len(precinct_pops) = votes_fractions.shape[1]
        """
        )

    if len(precinct_pops) != group_fractions.shape[1]:
        raise ValueError(
            """votes_fractions should have shape: r x num_precincts.
        In particular, it is required that len(precinct_pops) = group_fractions.shape[1]
        """
        )
    # check shapes of group_fractions, votes_fractions, and precinct_pops to make sure
    # number of precincts match (the last dimension of each)
    if not (
        len(votes_fractions[0]) == len(group_fractions[0])
        and len(group_fractions[0]) == len(precinct_pops)
    ):
        raise ValueError(
            """Mismatching num_precincts in input shapes. Inputs should have shape: \n
        votes_fraction shape: r x num_precincts \n
        group_fractions shape: c x num_precincts \n
        precinct_pops length: num_precincts
        """
        )



"""Helpers for managing data files."""

from dataclasses import dataclass

import pandas as pd

__all__ = ["Datasets"]


@dataclass
class _DataSet:
    """Class to hold datasets and related information.

    TODO: Add description, provenance, and other metadata here.
    """

    url: str

    def to_dataframe(self) -> pd.DataFrame:
        """Materialize as a pandas dataframe."""
        return pd.read_csv(self.url)


class Datasets:  # pylint: disable=too-few-public-methods
    """Available datasets related to ecological inference.

    These support examples in the library. Please open an issue or pull request if you would
    like to see other specific examples, or have questions about these."""

    Santa_Clara = _DataSet(
        "https://raw.githubusercontent.com/gerrymandr/ei-app/master/santaClara.csv"
    )
    Waterbury = _DataSet("https://raw.githubusercontent.com/gerrymandr/ei-app/master/waterbury.csv")



"""Test Goodmans Ecological Regression."""

# pylint:disable=redefined-outer-name
import numpy as np
import pytest

from test_plot_utils import example_two_by_two_data  # pylint:disable=unused-import
from pyei.goodmans_er import GoodmansER, GoodmansERBayes


@pytest.fixture
def group_and_vote_fractions():
    """Sample group and vote fractions, where every member of the demographic
    group votes for the given candidate and every non-member of the
    demographic group does not vote for the given candidate.
    """
    group_share = np.array([0, 0.2, 0.4, 0.6, 0.8])
    vote_share = np.array([0, 0.2, 0.4, 0.6, 0.8])
    return group_share, vote_share


@pytest.fixture
def group_and_vote_fractions_with_pop():
    """Sample group and vote fractions, where every member of the demographic
    group votes for the given candidate and 10% of the demographic group's
    complement supports the given candidate (i.e., slope = 1, intercept = 0.1),
    with an exception of one precinct.
    All precincts have population 1000, except one precinct that has population
    1 and does not follow the above formula.
    """
    group_share = np.array([0, 0.2, 0.4, 0.6, 0.8, 0.9])
    vote_share = np.array([0.1, 0.3, 0.5, 0.7, 0.9, 0.9])
    populations = np.array([1000, 1000, 1000, 1000, 1000, 1])
    return group_share, vote_share, populations


@pytest.fixture
def goodmans_er_bayes_examples(example_two_by_two_data):  # pylint: disable=redefined-outer-name
    """Run Bayesian Goodman's ER"""
    ex = example_two_by_two_data
    bayes_goodman_ei_weighted = GoodmansERBayes("goodman_er_bayes", weighted_by_pop=True, sigma=1)
    bayes_goodman_ei_weighted.fit(
        ex["group_fractions"],
        ex["votes_fractions"],
        ex["precint_pops"],
        demographic_group_name=ex["demographic_group_name"],
        candidate_name=ex["candidate_name"],
        tune=2000,
    )

    bayes_goodman_ei_unweighted = GoodmansERBayes(
        "goodman_er_bayes", weighted_by_pop=False, sigma=1
    )
    bayes_goodman_ei_unweighted.fit(
        ex["group_fractions"],
        ex["votes_fractions"],
        ex["precint_pops"],
        demographic_group_name=ex["demographic_group_name"],
        candidate_name=ex["candidate_name"],
    )
    return {
        "bayes_goodman_ei_weighted": bayes_goodman_ei_weighted,
        "bayes_goodman_ei_unweighted": bayes_goodman_ei_unweighted,
    }


def test_fit(group_and_vote_fractions):
    model = GoodmansER()
    group_share, vote_share = group_and_vote_fractions
    model.fit(group_share, vote_share)
    np.testing.assert_almost_equal(model.intercept_, 0)
    np.testing.assert_almost_equal(model.slope_, 1)


def test_weighted_fit(group_and_vote_fractions_with_pop):
    model = GoodmansER(is_weighted_regression=True)
    group_share, vote_share, pops = group_and_vote_fractions_with_pop
    model.fit(group_share, vote_share, pops)
    np.testing.assert_almost_equal(model.intercept_, 0.1, decimal=3)
    np.testing.assert_almost_equal(model.slope_, 1, decimal=3)


def test_summary():
    model = GoodmansER()
    model.demographic_group_name = "Trees"
    model.candidate_name = "Lorax"
    model.voting_prefs_est_ = 1.0
    model.voting_prefs_complement_est_ = 0.0
    expected_summary = """Goodmans ER
        Est. fraction of Trees
        voters who voted for Lorax is
        1.000
        Est. fraction of non- Trees
        voters who voted for Lorax is
        0.000
        """
    assert model.summary() == expected_summary
    model.is_weighted_regression = True
    expected_summary = """Goodmans ER, weighted by population
        Est. fraction of Trees
        voters who voted for Lorax is
        1.000
        Est. fraction of non- Trees
        voters who voted for Lorax is
        0.000
        """
    assert model.summary() == expected_summary


def test_plot(group_and_vote_fractions):
    model = GoodmansER()
    group_share, vote_share = group_and_vote_fractions
    model.fit(group_share, vote_share)
    _, ax = model.plot()
    x_plot, y_plot = ax.lines[0].get_xydata().T
    np.testing.assert_allclose(x_plot, y_plot, atol=1e-10)
    assert (0.0, 1.0) == ax.get_xlim()
    assert (0.0, 1.0) == ax.get_ylim()


def test_goodman_er_bayes_posterior_means(goodmans_er_bayes_examples):
    goodmans_er_bayes_weighted = goodmans_er_bayes_examples["bayes_goodman_ei_weighted"]
    np.testing.assert_almost_equal(
        goodmans_er_bayes_weighted.sampled_voting_prefs[0].mean(), 0.840, decimal=2
    )
    np.testing.assert_almost_equal(
        goodmans_er_bayes_weighted.sampled_voting_prefs[1].mean(), 0.240, decimal=2
    )

    goodmans_er_bayes_unweighted = goodmans_er_bayes_examples["bayes_goodman_ei_unweighted"]
    np.testing.assert_almost_equal(
        goodmans_er_bayes_unweighted.sampled_voting_prefs[0].mean(), 0.835, decimal=2
    )
    np.testing.assert_almost_equal(
        goodmans_er_bayes_unweighted.sampled_voting_prefs[1].mean(), 0.244, decimal=2
    )


def test_goodman_er_bayes_bounds(goodmans_er_bayes_examples):
    goodmans_er_bayes_example = goodmans_er_bayes_examples["bayes_goodman_ei_weighted"]
    (
        _,
        _,
        lower_bounds,
        upper_bounds,
    ) = goodmans_er_bayes_example.compute_credible_int_for_line()

    assert all(upper_bounds) <= 1.0
    assert all(upper_bounds) >= 0

    assert all(lower_bounds) <= 1.0
    assert all(lower_bounds) >= 0


def test_goodman_er_bayes_plot(goodmans_er_bayes_examples):
    ax = goodmans_er_bayes_examples["bayes_goodman_ei_weighted"].plot()
    assert ax is not None



import codecs
import os
import re

import setuptools
from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop


PROJECT_ROOT = os.path.dirname(os.path.realpath(__file__))
REQUIREMENTS_FILE = os.path.join(PROJECT_ROOT, "requirements.txt")
README_FILE = os.path.join(PROJECT_ROOT, "README.md")
VERSION_FILE = os.path.join(PROJECT_ROOT, "pyei", "__init__.py")


def get_requirements():
    # TODO: The gh-action to upload to PyPI doesn't include the requirements.txt
    # so we're just copying them here as a backup. This is super brittle!
    try:
        with codecs.open(REQUIREMENTS_FILE, "r", encoding="utf-8") as buff:
            return buff.read().splitlines()
    except:
        return """pymc >= 5.18.0
arviz
scikit-learn
matplotlib
pandas
seaborn
graphviz
numpy
jax
numpyro
jaxlib
numba
netCDF4""".splitlines()


def get_long_description():
    with codecs.open(README_FILE, "r", encoding="utf-8") as buff:
        return buff.read()


def get_version():
    lines = open(VERSION_FILE, "r", encoding="utf-8").readlines()
    version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in lines:
        mo = re.search(version_regex, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError("Unable to find version in %s." % (VERSION_FILE,))


setup(
    name="pyei",
    version=get_version(),
    description="",
    author="Metric Geometry and Gerrymandering Group",
    url="https://github.com/mggg/ecological-inference",
    packages=find_packages(),
    install_requires=get_requirements(),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    include_package_data=True,
)



# pylint: disable-all
"""
Port of R code for Noncentral Hypergeometric distribution
adapted from R code published in conjunction with:
Liao, J.G. And Rosen, O. (2001) Fast and Stable Algorithms for Computing and
Sampling from the Noncentral Hypergeometric Distribution.  The American
Statistician 55, 366-369.

Used in Greiner-Quinn method Gibbs sampler
"""
import math
import numpy as np
from numba import jit


@jit
def r_function(n1, n2, m1, psi, i):
    """
    The function r defined in Liao and Rosen 2001
    """
    return (n1 - i + 1) * (m1 - i + 1) / (i * (n2 - m1 + i)) * psi


@jit
def sample_low_to_high(lower, ran, pi, shift, uu):
    for i in range(lower, uu + 1):
        if ran <= pi[i + shift]:
            return i
        ran = ran - pi[i + shift]


@jit
def sample_high_to_low(upper, ran, pi, shift, ll):
    for i in range(upper, ll - 1, -1):
        if ran <= pi[i + shift]:
            return i
        ran = ran - pi[i + shift]


@jit
def non_central_hypergeometric_sample(n1, n2, m1, psi):
    """Allows for sampling from noncentralhypergeometric distribution
    Following the methods of Liao and Rosen, 2001

    If
    y1 ~ Binom(n1, pi1)
    y2 ~ Binom(n2, pi2)
    psi = pi1 (1-pi2) / pi2 (1-pi1)

    this the nchg distribution governs with parameters n1, n2, pis, m1
    is the distribution for y1 | y1 + y2 = m1
    """

    ll = max(0, m1 - n2)
    uu = min(n1, m1)

    a = psi - 1
    b = -((m1 + n1 + 2) * psi + n2 - m1)
    c = psi * (n1 + 1) * (m1 + 1)
    q = -(b + np.sign(b) * np.sqrt(b * b - 4 * a * c)) / 2
    mode_candidate = math.trunc(c / q)
    if (uu >= mode_candidate) and (ll <= mode_candidate):
        mode = mode_candidate
    else:
        mode = math.trunc(q / a)

    # calculate density
    pi = np.ones(uu - ll + 1, dtype=np.float32)

    if mode < uu:
        r = r_function(n1, n2, m1, psi, np.arange(mode + 1, uu + 1))
        pi[(mode + 1 - ll) : (uu - ll + 1)] = np.cumprod(r)

    if mode > ll:
        r = 1 / r_function(
            n1,
            n2,
            m1,
            psi,
            np.flip(np.arange(ll + 1, mode + 1)),
        )
        pi[0 : (mode - ll)] = np.flip(np.cumprod(r))
    density = pi / pi.sum()

    ran = np.random.uniform(0, 1)
    pi = density

    if mode == ll:
        return sample_low_to_high(ll, ran, pi, -ll, uu)
    if mode == uu:
        return sample_high_to_low(uu, ran, pi, -ll, ll)
    if ran < pi[mode - ll]:
        return mode
    ran = ran - pi[mode - ll]
    lower = mode - 1
    upper = mode + 1

    while True:
        if pi[upper - ll] >= pi[lower - ll]:
            if ran < pi[upper - ll]:
                return upper
            ran = ran - pi[upper - ll]
            if upper == uu:
                samp = sample_high_to_low(lower, ran, pi, -ll, ll)
                return samp
            upper = upper + 1

        else:
            if ran < pi[lower - ll]:
                return lower
            ran = ran - pi[lower - ll]
            if lower == ll:
                samp = sample_low_to_high(upper, ran, pi, -ll, uu)
                return samp
            lower = lower - 1



"""Tests for data.py"""

import pytest

from pyei import data


@pytest.fixture
def expected_data_sets():
    """Hardcode the names of existing datasets.

    We could also use `dir` or `vars`, but this is quicker for now...
    """
    return ("Santa_Clara", "Waterbury")


def test_data_sets_exist(expected_data_sets):  # pylint: disable=redefined-outer-name
    for data_set in expected_data_sets:
        assert hasattr(data.Datasets, data_set)
